import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.io.PrintStream;
import java.lang.reflect.Method;

/**
 * <pre>   
 * Node -- Class defining the data structures and code for the
 * protocol stack for a node participating in the fishnet

 * This code is under student control; we provide a baseline
 * "hello world" example just to show how it interfaces with the FishNet classes.
 *
 * NOTE: per-node global variables are prohibited.
 *  In simulation mode, we execute many nodes in parallel, and thus
 *  they would share any globals.  Please put per-node state in class Node
 *  instead.
 *
 * In other words, all the code defined here should be generic -- it shouldn't
 * matter to this code whether it is executing in a simulation or emulation
 * 
 * This code must be written as a state machine -- each upcall must do its work and return so that 
 * other upcalls can be delivered
 * </pre>   
 */
public class Node {	
	// Note that the current implementation works only if PingTimeout < NeighborProbeInterval
	private final long PingTimeout = 10000;  // Timeout pings in 10 seconds
	private final long NeighborProbeInterval = 30000; // Interval at which to update neighbors
	private final String neighborPrefix = "neighbor ping ";
    
    private Manager manager;
    private int addr;
    private int seq;
    private int lastNeighborPingSeq;
    private ArrayList<PingRequest> pings; // To store PingRequests.
    //TODO: occasionally purge old information from seenPkts?
    private HashSet<SrcSeqPair> seenPkts; // records src,seq pairs of seen packets
    private ArrayList<Integer> neighbors; // the current stable list of neighbors
    private ArrayList<Integer> newNeighbors; // the set of new neighbors being constructed

    /**
     * Create a new node
     * @param manager The manager that is managing Fishnet
     * @param addr The address of this node
     */
    public Node(Manager manager, int addr) {
	this.manager = manager;
	this.addr = addr;
	this.seq = 0;
	this.lastNeighborPingSeq = 0;
	this.pings = new ArrayList<PingRequest>();
	this.seenPkts = new HashSet<SrcSeqPair>();
	this.neighbors = new ArrayList<Integer>();
	this.newNeighbors = new ArrayList<Integer>();
    }

    /**
     * Called by the manager to start this node up.
     */
    public void start() {
	logOutput("started");
	this.addTimer(PingTimeout, "pingTimedOut");
	this.sendNeighborPing();
    }

    /**
     * Called by the manager when a packet has arrived for this node.
     * @param from The address of the node that has sent this packet
     * @param msg The serialized form of the packet.
     */
    public void onReceive(Integer from, byte[] msg) {	
	Packet packet = Packet.unpack(msg);
	if(packet == null) {
	    logError("Unable to unpack message: " + Utility.byteArrayToString(msg) + " Received from " + from);
	    return;
	}

	// If the packet is destined for us, process it
	if( packet.getDest() == this.addr || packet.getDest() == Packet.BROADCAST_ADDRESS ){
		logOutput("received packet from " + from);
		this.receivePacket(from.intValue(), packet);		
	}
	
	// If the packet is destined elsewhere forward it
	if( packet.getDest() != this.addr ){
		this.forwardPacket( from.intValue(), packet );
	}
    }

    /**
     * Called by the manager when there is a command for this node from the user.
     * Command can be input either from keyboard or file.
     * For now sending messages as ping to a neighbor.
     * @param command The command for this node
     */
    public void onCommand(String command) {
	if(this.matchPingCommand(command)) {
	    return;
	}else if(this.matchNeighborCommand(command)){
		return;
	}
	
	logError("Unrecognized command: " + command);
    }

    /**
     * Callback method given to manager to invoke when a timer fires.
     */
    public void pingTimedOut() {
	Iterator<PingRequest> iter = this.pings.iterator();
	while(iter.hasNext()) {
	    PingRequest pingRequest = (PingRequest)iter.next();
	    if((pingRequest.getTimeSent() + PingTimeout) < this.manager.now()) {
		try {
		    logOutput("Timing out ping: " + pingRequest);
		    iter.remove();
		}catch(Exception e) {
		    logError("Exception occured while trying to remove an element from ArrayList pings.  Exception: " + e);
		    return;
		}
	    }
	}
	this.addTimer(PingTimeout, "pingTimedOut");
    }
    
    /**
     * Called when the pings sent to look for new neighbors should have timed out.
     * Signals the end of this round of neighbor discovery, replacing the old neighbor list
     * with the new one.
     */
    public void neighborTimedOut(){
    	this.neighbors = this.newNeighbors;
    	this.logOutput( "Current neighbors: " + this.neighbors );
    }

    private boolean matchPingCommand(String command) {
	int index = command.indexOf(" ");
	if(index == -1) {
	    return false;
	}
	try {
	    int destAddr = Integer.parseInt(command.substring(0, index));
	    String message = command.substring(index+1);
	    Packet packet = new Packet(destAddr, this.addr, Packet.MAX_TTL, Protocol.PING_PKT, 
	    				this.seq++, Utility.stringToByteArray(message));

	    this.floodSend(packet);
	    this.pings.add(new PingRequest(destAddr, Utility.stringToByteArray(message), this.manager.now()));
	    return true;
	}catch(Exception e) {
	    logError("Exception: " + e);
	}
		
	return false;
    }
    
    private boolean matchNeighborCommand(String command) {
    	if( !command.startsWith( "neighbor" ) ){
    		return false;
    	}
    	
    	this.logOutput( "Neighbors: " + neighbors );
    	return true;
    }
    
    /**
     * Decrement the TTL and forward this packet if it's greater than 0. 
     * Currently forwards by flooding to the broadcast address.
     * @param from The address this packet was received from (currently ignored).
     * @param packet The relevant packet
     */
    private void forwardPacket(int from, Packet packet){
    	
    	if( seenPkts.contains( new SrcSeqPair( packet.getSrc(), packet.getSeq() ) ) ){
    		// drop the packet because we've seen it
    		return;
    	}
    	
    	// store that we've seen it
    	seenPkts.add( new SrcSeqPair( packet.getSrc(), packet.getSeq() ) );

    	// decrement the TTL and forward if it's greater than 0
		packet.setTTL( packet.getTTL() - 1 );
		if( packet.getTTL() > 0 ){
			// if the destination is a neighbor, send directly, otherwise flood
			floodSend( packet );
    	}
	}
    
    /**
     * Called periodically to refresh the neighbor list. Sends a ping with a special payload
     * and TTL so that the reply will be recognizable and the ping will not travel beyond
     * immediate neighbors.
     */
    public void sendNeighborPing(){
    	// special payload
    	String payload = this.neighborPrefix + this.seq;
    	
    	// used to differentiate between neighbor discovery "rounds"
    	this.lastNeighborPingSeq = this.seq;

    	// clear the list of newNeighbors
    	this.newNeighbors = new ArrayList<Integer>();

    	// build and send the special neighbor ping packets
    	Packet packet = new Packet( Packet.BROADCAST_ADDRESS, this.addr, 1, Protocol.PING_PKT, this.seq++, Utility.stringToByteArray( payload ) );
    	floodSend( packet );
    	
    	// schedule a timer to end this round and timer to start the next
    	this.addTimer( this.NeighborProbeInterval, "sendNeighborPing" );
    	this.addTimer( this.PingTimeout, "neighborTimedOut" );
    }

    private void receivePacket(int from, Packet packet) {

	switch(packet.getProtocol()) {

	case Protocol.PING_PKT:
	    this.receivePing(packet);
	    break;
	    
	case Protocol.PING_REPLY_PKT:
	    this.receivePingReply(packet);
	    break;

	default:
	    logError("Packet with unknown protocol received. Protocol: " + packet.getProtocol());	    
	}
    }

    private void receivePing(Packet packet) {
	logOutput("Received Ping from " + packet.getSrc() + " with message: " + Utility.byteArrayToString(packet.getPayload()));
	
	try {
	    Packet reply = new Packet(packet.getSrc(), this.addr, Packet.MAX_TTL, Protocol.PING_REPLY_PKT, seq++, packet.getPayload());
	    if( Utility.byteArrayToString( packet.getPayload() ).startsWith( this.neighborPrefix ) ){
	    	// It's ostensibly from a neighbor, so we should be able to send directly
	    	this.send( packet.getSrc(), reply );
	    }else{
	    	this.floodSend(reply);
	    }
	}catch(IllegalArgumentException e) {
	    logError("Exception while trying to send a Ping Reply. Exception: " + e);
	}
    }

    // Check that ping reply matches what was sent
    private void receivePingReply(Packet packet) {
    
    if( Utility.byteArrayToString( packet.getPayload() ).startsWith( this.neighborPrefix ) ){
    	receiveNeighborPingReply( packet );
    	return;
    }
    
	Iterator<PingRequest> iter = this.pings.iterator();
	String payload = Utility.byteArrayToString(packet.getPayload());
	while(iter.hasNext()) {
	    PingRequest pingRequest = (PingRequest)iter.next();
	    if( (pingRequest.getDestAddr() == packet.getSrc()) &&
		( Utility.byteArrayToString(pingRequest.getMsg()).equals(payload))) {
		
		logOutput("Got Ping Reply from " + packet.getSrc() + ": " + payload);
		try {
		    iter.remove();
		}catch(Exception e) {
		    logError("Exception occured while trying to remove an element from ArrayList pings while processing Ping Reply.  Exception: " + e);
		}
		return;
	    }
	}
	logError("Unexpected Ping Reply from " + packet.getSrc() + ": " + payload);
    }
    
    /**
     * Called when we receive a ping response from our neighbor discovery. Adds this neighbor
     * to the new list of neighbors.
     * @param packet the received packet
     */
    private void receiveNeighborPingReply( Packet packet ){
    	// extract the sequence number from the packet
    	String seqStr = Utility.byteArrayToString( packet.getPayload() ).substring( 14 );
    	int seqNo = Integer.parseInt( seqStr );
    	
    	// if it's from a new neighbor for this neighbor discovery round, deal with it
    	if( seqNo == this.lastNeighborPingSeq && !this.newNeighbors.contains( packet.getSrc() ) ){
    		this.newNeighbors.add( packet.getSrc() );
    		this.logOutput( "Addding " + packet.getSrc() + " as a neighbor" );
    	}else{
    		this.logError( "Received ping from " + packet.getSrc() + " in the wrong round; Got " + 
    				seqNo + ", expected " + this.lastNeighborPingSeq );
    	}
    }
    
    private void send(int destAddr, Packet packet) {
	try {
	    this.manager.sendPkt(this.addr, destAddr, packet.pack());
	}catch(IllegalArgumentException e) {
	    logError("Exception: " + e);
	}
    }
    
    /**
     * Same as send, but sends to the broadcast address if the destination is not a neighbor.
     * @param packet the packet to be sent
     */
    private void floodSend( Packet packet ){
		if( this.neighbors.contains( packet.getDest() ) ){
			send( packet.getDest(), packet );
		}else{
			this.send( Packet.BROADCAST_ADDRESS, packet );
		}
    }

    // Adds a timer, to fire in deltaT milliseconds, with a callback to a public function of this class that takes no parameters
    private void addTimer(long deltaT, String methodName) {
	try {
	    Method method = Callback.getMethod(methodName, this, null);
	    Callback cb = new Callback(method, this, null);
	    this.manager.addTimer(this.addr, deltaT, cb);
	}catch(Exception e) {
	    logError("Failed to add timer callback. Method Name: " + methodName +
		     "\nException: " + e);
	}
    }

    private void logError(String output) {
	this.log(output, System.err);
    }

    private void logOutput(String output) {
	this.log(output, System.out);
    }

    private void log(String output, PrintStream stream) {
	stream.println("Node " + this.addr + ": " + output);
    }

    /**
     * A class which stores a pair of integers, src and seq, representing a unique
     * packet ID.
     */
	private static class SrcSeqPair{
		int src;
		int seq;
		
		public SrcSeqPair( int inSrc, int inSeq ){
			this.src = inSrc;
			this.seq = inSeq;
		}
		
		public int getSrc(){
			return src;
		}
		
		public int getSeq(){
			return seq;
		}
		
		@Override
		public boolean equals( Object o ){
			if( o instanceof SrcSeqPair ){
				SrcSeqPair ssp = (SrcSeqPair)o;
				return this.src == ssp.src && this.seq == ssp.seq;
			}
			return false;
		}
		
		@Override
		public int hashCode(){
			return src ^ seq;
		}
	}
}
